# views.py
#
# Copyright (c) 2024 - 2025 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import logging
from typing import TYPE_CHECKING, ClassVar

# from django.shortcuts import render
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    SuspiciousOperation,
)
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    FormView,
    ListView,
    TemplateView,
    UpdateView,
)
from django.views.generic.detail import SingleObjectMixin

if TYPE_CHECKING:
    from django.db.models import QuerySet

from podcast_analyzer.forms import AnalysisGroupForm, PersonMergeForm
from podcast_analyzer.models import AnalysisGroup, Episode, Person, Podcast

# Create your views here.

logger = logging.getLogger(__name__)


def messaging_enabled() -> bool:
    """Returns a bool representing if django's messages are enabled."""
    return "django.contrib.messages" in settings.INSTALLED_APPS


class SelectPrefetchRelatedMixin:
    """
    Adds two class variables:
        prefetch_related: list[str]
        select_related: list[str]
    If they are populated, it will use them in the queryset.
    """

    model = None
    queryset = None
    ordering = None
    prefetch_related: ClassVar[list[str] | None] = None
    select_related: ClassVar[list[str] | None] = None

    def get_queryset(self):
        if not self.queryset:
            if self.model:
                qs = self.model._default_manager.all()
            else:  # no cov
                msg = "View does not have a queryset or model defined!"
                raise ImproperlyConfigured(msg)
        else:
            qs = self.queryset
        if self.select_related:
            qs = qs.select_related(*self.select_related)
        if self.prefetch_related:
            qs = qs.prefetch_related(*self.prefetch_related)
        if self.ordering:
            if isinstance(self.ordering, str):
                qs = qs.order_by(self.ordering)  # no cov, we're just mimicking Django
            else:
                qs = qs.order_by(*self.ordering) # type: ignore
        return qs


class AppEntryView(LoginRequiredMixin, TemplateView):
    """The default view for the root of the application."""

    template_name = "podcast_analyzer/app_entry.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["podcast_count"] = Podcast.objects.count()
        context["people_count"] = Person.objects.count()
        context["analysis_group_count"] = AnalysisGroup.objects.count()
        context["episode_count"] = Episode.objects.count()
        return context


class PodcastListView(LoginRequiredMixin, SelectPrefetchRelatedMixin, ListView):
    """
    View all podcasts in a list.
    """

    model = Podcast
    ordering = ["title"]
    paginate_by = 25
    prefetch_related = ["episodes", "seasons", "analysis_group"]


class PodcastDetailView(LoginRequiredMixin, SelectPrefetchRelatedMixin, DetailView):
    """
    A view to see a given podcasts data.
    """

    model = Podcast
    pk_url_kwarg = "id"
    context_object_name = "podcast"
    prefetch_related = ["episodes", "seasons", "analysis_group"]


class PodcastCreateView(LoginRequiredMixin, CreateView):
    """
    Provides a form to create a podcast.
    """

    model = Podcast
    fields: ClassVar[list[str]] = ["title", "rss_feed"]  # type: ignore

    def form_invalid(self, form):
        if messaging_enabled():
            messages.error(
                self.request, _("Podcast could not be created. See errors below.")
            )
        return super().form_invalid(form)

    def get_success_url(self):
        if messaging_enabled():
            messages.success(self.request, _("Podcast created"))
        return super().get_success_url()


class PodcastUpdateView(LoginRequiredMixin, UpdateView):
    """
    View for updating a podcast record.
    """

    model = Podcast
    pk_url_kwarg = "id"
    fields: ClassVar[list[str]] = [  # type: ignore
        "title",
        "rss_feed",
        "site_url",
        "podcast_cover_art_url",
        "release_frequency",
        "probable_feed_host",
        "tags",
        "analysis_group",
    ]
    context_object_name = "podcast"

    def form_invalid(self, form):
        if messaging_enabled():
            messages.error(
                self.request, _("Podcast could not be updated. See errors below.")
            )
        return super().form_invalid(form)

    def get_success_url(self):
        if messaging_enabled():
            messages.success(self.request, _("Podcast updated"))
        return super().get_success_url()


class PodcastDeleteView(LoginRequiredMixin, SelectPrefetchRelatedMixin, DeleteView):
    """
    For deleting a podcast record.
    """

    model = Podcast
    context_object_name = "podcast"
    pk_url_kwarg = "id"
    object: Podcast
    prefetch_related = ["episodes", "seasons", "analysis_group"]

    def get_success_url(self):
        if messaging_enabled():
            messages.success(self.request, _("Podcast deleted"))
        return reverse_lazy("podcast_analyzer:podcast-list")


class PersonListView(LoginRequiredMixin, SelectPrefetchRelatedMixin, ListView):
    """View all people in a list"""

    model = Person
    pk_url_kwarg = "id"
    context_object_name = "people"
    paginate_by = 25
    ordering = ["name"]
    prefetch_related = ["hosted_episodes", "guest_appearances"]
    queryset = Person.objects.filter(merged_into__isnull=True)


class PersonDetailView(LoginRequiredMixin, SelectPrefetchRelatedMixin, DetailView):
    """View a specific person's details."""

    model = Person
    pk_url_kwarg = "id"
    context_object_name = "person"
    prefetch_related = ["hosted_episodes", "guest_appearances"]


class PersonUpdateView(LoginRequiredMixin, UpdateView):
    """Edit a given person's details."""

    model = Person
    fields: ClassVar[list[str]] = ["name", "url", "img_url"]
    context_object_name = "person"
    pk_url_kwarg = "id"

    def form_invalid(self, form):
        if messaging_enabled():
            messages.error(
                self.request, _("Person could not be updated. See errors below.")
            )
        return super().form_invalid(form)

    def get_success_url(self):
        if messaging_enabled():
            messages.success(self.request, _("Person updated"))
        return super().get_success_url()


class PersonDeleteView(LoginRequiredMixin, SelectPrefetchRelatedMixin, DeleteView):
    """Delete a person from the database."""

    model = Person
    pk_url_kwarg = "id"
    context_object_name = "person"
    prefetch_related = ["hosted_episodes", "guest_appearances"]
    select_related = ["merged_into"]
    object: Person

    def get_success_url(self):
        if messaging_enabled():
            messages.success(self.request, _("Person deleted"))
        return reverse_lazy("podcast_analyzer:person-list")


class PersonMergeListView(LoginRequiredMixin, SelectPrefetchRelatedMixin, ListView):
    """Show a list of people that this record could be merged with."""

    model = Person
    context_object_name = "merge_targets"
    prefetch_related = ["hosted_episodes", "guest_appearances"]
    select_related = ["merged_into"]
    template_name = "podcast_analyzer/person_merge_list.html"
    paginate_by = 20
    source_person: Person

    def get_queryset(self):
        person_id = self.kwargs.get("id", None)
        self.source_person = get_object_or_404(Person, pk=person_id)
        return Person.objects.filter(merged_into__isnull=True).exclude(
            id=self.source_person.id
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["person"] = self.source_person
        return context


class PersonMergeView(LoginRequiredMixin, SingleObjectMixin, FormView):
    """
    Show a confirmation screen so that the user can double-check what will be changed.
    Then perform the merge on a post.
    """

    model = Person
    pk_url_kwarg = "id"
    context_object_name = "source"
    destination: Person
    template_name = "podcast_analyzer/person_confirm_merge.html"
    form_class = PersonMergeForm
    object: Person

    def form_valid(self, form):
        source = form.cleaned_data.get("source_person")
        destination = form.cleaned_data.get("destination_person")
        if destination != self.destination or source != self.object:
            msg = "It looks like you are trying to tamper with the merge form."
            raise SuspiciousOperation(msg)
        records_updated = Person.merge_person(
            source_person=source, destination_person=destination
        )
        if messaging_enabled():
            messages.success(
                self.request,
                f"Merged successfully and updated {records_updated} related records.",
            )
        return HttpResponseRedirect(destination.get_absolute_url())

    def get_form_kwargs(self):
        if not hasattr(self, "object"):
            obj = self.get_object()
        else:
            obj = self.object
        self.initial = {"source_person": obj, "destination_person": self.destination}
        return super().get_form_kwargs()

    def get_queryset(self):
        qs = (
            Person.objects.filter(merged_into__isnull=True)
            .prefetch_related("hosted_episodes", "guest_appearances")
            .select_related("merged_into")
        )
        destination_id = self.kwargs.get("destination_id")
        try:
            self.destination = qs.get(id=destination_id)
        except ObjectDoesNotExist as odne:
            raise Http404 from odne
        return qs

    def get_object(self, queryset=None):
        self.object = super().get_object(queryset=queryset)  # type: ignore
        return self.object

    def get_context_data(self, **kwargs):
        if not hasattr(self, "object"):
            self.get_object()
        context = super().get_context_data(**kwargs)
        context["destination"] = self.destination
        context["conflict_data"] = self.object.get_potential_merge_conflicts(
            self.destination
        )
        return context


class PodcastEpisodeDescendantMixin(SelectPrefetchRelatedMixin):
    """
    Allows the episode queryset to be filtered by the podcast it comes from.
    """

    podcast: Podcast | None = None
    if TYPE_CHECKING:
        queryset: QuerySet[Episode] | None

    def get_queryset(self):
        podcast_id = self.kwargs.get("podcast_id", None)  # type: ignore
        if not podcast_id:  # no cov
            raise Http404
        try:
            self.podcast = Podcast.objects.get(pk=podcast_id)
        except ObjectDoesNotExist as ode:  # no cov
            raise Http404 from ode
        self.queryset = Episode.objects.filter(podcast=self.podcast)
        return super().get_queryset()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)  # type: ignore
        context["podcast"] = self.podcast
        return context


class EpisodeListView(LoginRequiredMixin, PodcastEpisodeDescendantMixin, ListView):
    """View all episodes for given podcast"""

    model = Episode
    pk_url_kwarg = "id"
    context_object_name = "episodes"
    paginate_by = 50
    ordering = ["-release_datetime", "-ep_num"]
    select_related = ["podcast", "season"]


class EpisodeDetailView(LoginRequiredMixin, PodcastEpisodeDescendantMixin, DetailView):
    """View a given episode's details."""

    model = Episode
    pk_url_kwarg = "id"
    context_object_name = "episode"
    select_related = ["podcast", "season"]
    prefetch_related = [
        "analysis_group",
        "hosts_detected_from_feed",
        "guests_detected_from_feed",
    ]


class EpisodeUpdateView(LoginRequiredMixin, PodcastEpisodeDescendantMixin, UpdateView):
    """Edit a given episode's details."""

    model = Episode
    pk_url_kwarg = "id"
    context_object_name = "episode"
    fields = [
        "title",
        "ep_num",
        "ep_type",
        "analysis_group",
        "hosts_detected_from_feed",
        "guests_detected_from_feed",
    ]

    def get_success_url(self):
        if messaging_enabled():
            messages.success(self.request, _("Episode updated"))
        return super().get_success_url()

    def form_invalid(self, form):  # no cov
        if messaging_enabled():
            messages.error(
                self.request, _("Episode could not be updated. See errors below.")
            )
        return super().form_invalid(form)


class EpisodeDeleteView(LoginRequiredMixin, PodcastEpisodeDescendantMixin, DeleteView):
    """Delete a given episode."""

    model = Episode
    pk_url_kwarg = "id"
    context_object_name = "episode"
    object: Episode
    select_related = ["podcast", "season"]
    prefetch_related = [
        "analysis_group",
        "hosts_detected_from_feed",
        "guests_detected_from_feed",
    ]
    podcast: Podcast

    def get_success_url(self):
        if messaging_enabled():
            messages.success(self.request, _("Episode deleted"))
        return reverse_lazy(
            "podcast_analyzer:episode-list", kwargs={"podcast_id": self.podcast.pk}
        )


class AnalysisGroupListView(LoginRequiredMixin, SelectPrefetchRelatedMixin, ListView):
    """List of all analysis groups."""

    model = AnalysisGroup
    context_object_name = "groups"
    paginate_by = 25
    prefetch_related = ["podcasts", "seasons", "episodes"]
    ordering = ["name", "-created"]


class AnalysisGroupDetailView(
    LoginRequiredMixin, SelectPrefetchRelatedMixin, DetailView
):
    """Details for a given analysis group."""

    model = AnalysisGroup
    pk_url_kwarg = "id"
    context_object_name = "group"
    prefetch_related = ["podcasts", "seasons", "episodes"]


class AnalysisGroupUpdateView(
    LoginRequiredMixin, SelectPrefetchRelatedMixin, UpdateView
):
    """Update a given analysis group."""

    model = AnalysisGroup
    pk_url_kwarg = "id"
    context_object_name = "group"
    form_class = AnalysisGroupForm
    prefetch_related = ["podcasts", "seasons", "episodes"]
    object: AnalysisGroup

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["initial"]["podcasts"] = self.object.podcasts.all()
        kwargs["initial"]["seasons"] = self.object.seasons.all()
        kwargs["initial"]["episodes"] = self.object.episodes.all()
        return kwargs

    def form_valid(self, form):
        self.object = form.save()
        self.object.podcasts.set(form.cleaned_data["podcasts"])
        self.object.seasons.set(form.cleaned_data["seasons"])
        self.object.episodes.set(form.cleaned_data["episodes"])
        if messaging_enabled():
            messages.success(self.request, _("Analysis group updated"))
        return HttpResponseRedirect(self.get_success_url())


class AnalysisGroupCreateView(LoginRequiredMixin, CreateView):
    """Create a new analysis group."""

    model = AnalysisGroup
    form_class = AnalysisGroupForm
    object: AnalysisGroup

    def form_valid(self, form):
        self.object = AnalysisGroup.objects.create(name=form.cleaned_data["name"])
        logger.debug(
            f"Created new analysis group with name: {form.cleaned_data['name']}"
        )
        logger.debug(
            f"Adding {len(form.cleaned_data['podcasts'])} podcasts, "
            f"{len(form.cleaned_data['seasons'])} seasons and "
            f"{len(form.cleaned_data['episodes'])} episodes."
        )
        self.object.podcasts.set(form.cleaned_data["podcasts"])
        self.object.seasons.set(form.cleaned_data["seasons"])
        self.object.episodes.set(form.cleaned_data["episodes"])
        if messaging_enabled():
            messages.success(self.request, _("Analysis group created"))
        return HttpResponseRedirect(self.object.get_absolute_url())


class AnalysisGroupDeleteView(
    LoginRequiredMixin, SelectPrefetchRelatedMixin, DeleteView
):
    """Delete a given analysis group."""

    model = AnalysisGroup
    pk_url_kwarg = "id"
    context_object_name = "group"
    object: AnalysisGroup
    prefetch_related = ["podcasts", "seasons", "episodes"]

    def get_success_url(self):
        if messaging_enabled():
            messages.success(self.request, _("Analysis group deleted"))
        return reverse_lazy("podcast_analyzer:ag-list")


class TagPodcastListView(LoginRequiredMixin, SelectPrefetchRelatedMixin, ListView):
    """
    List the podcasts for the given tag.
    """

    model = Podcast
    slug_field = "tag_slug"
    context_object_name = "podcasts"
    paginate_by = 25
    prefetch_related = ["episodes", "seasons", "itunes_categories", "tags"]
    ordering = ["title", "-created"]
    tag: Podcast.tags.tag_model  # type: ignore

    def get_queryset(self):
        try:
            self.tag = Podcast.tags.tag_model.objects.get(
                slug=self.kwargs.get("tag_slug", None)
            )
        except ObjectDoesNotExist as odne:
            raise Http404 from odne
        self.queryset = Podcast.objects.filter(tags__in=[self.tag])
        return super().get_queryset()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["tag"] = self.tag
        return context



# test_admin.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""Tests for the admin actions."""

from io import BytesIO

import pytest
from django.contrib import admin, messages

from podcast_analyzer.admin import PodcastAdmin
from podcast_analyzer.models import Podcast

pytestmark = pytest.mark.django_db(transaction=True)


@pytest.mark.parametrize("expected_new", [0, 2, 4])
def test_get_new_episodes(
    rf,
    settings,
    user,
    httpx_mock,
    podcast_with_parsed_episodes,
    expected_new: int,
) -> None:
    with open("tests/data/podcast_rss_feed.xml", "rb") as f:
        datastream = BytesIO(f.read())
    settings.MESSAGE_STORAGE = "django.contrib.messages.storage.cookie.CookieStorage"
    if expected_new > 0:
        episodes = podcast_with_parsed_episodes.episodes.all().order_by(
            "-release_datetime"
        )[:expected_new]
        for episode in episodes:
            episode.delete()
    ep_count = podcast_with_parsed_episodes.episodes.count()
    podcast_queryset = Podcast.objects.filter(id=podcast_with_parsed_episodes.id)
    httpx_mock.add_response(
        url=podcast_with_parsed_episodes.rss_feed, content=datastream
    )
    admin_obj = PodcastAdmin(Podcast, admin.site)
    request = rf.get("/admin/podcast_analyzer/")
    request.user = user
    request._messages = messages.storage.default_storage(request)
    admin_obj.check_for_new_episodes(request, podcast_queryset)
    podcast_with_parsed_episodes.refresh_from_db()
    assert podcast_with_parsed_episodes.episodes.count() - ep_count == expected_new


def test_get_all_episodes(
    rf, settings, httpx_mock, user, podcast_with_parsed_episodes
) -> None:
    settings.MESSAGE_STORAGE = "django.contrib.messages.storage.cookie.CookieStorage"
    with open("tests/data/podcast_rss_feed.xml", "rb") as f:
        datastream = BytesIO(f.read())
    httpx_mock.add_response(
        url=podcast_with_parsed_episodes.rss_feed, content=datastream
    )
    latest_ep = podcast_with_parsed_episodes.episodes.latest("release_datetime")
    original_duration = latest_ep.itunes_duration
    latest_ep.itunes_duration = 2460
    latest_ep.save()
    podcast_queryset = Podcast.objects.filter(id=podcast_with_parsed_episodes.id)
    request = rf.get("/admin/podcast_analyzer/")
    request.user = user
    request._messages = messages.storage.default_storage(request)
    admin_obj = PodcastAdmin(Podcast, admin.site)
    admin_obj.refresh_all_episodes(request, podcast_queryset)
    latest_ep.refresh_from_db()
    assert latest_ep.itunes_duration == original_duration


def test_feed_update(
    rf, settings, httpx_mock, user, podcast_with_parsed_episodes
) -> None:
    settings.MESSAGE_STORAGE = "django.contrib.messages.storage.cookie.CookieStorage"
    with open("tests/data/podcast_rss_feed.xml", "rb") as f:
        datastream = BytesIO(f.read())
    httpx_mock.add_response(
        url=podcast_with_parsed_episodes.rss_feed, content=datastream
    )
    latest_ep = podcast_with_parsed_episodes.episodes.latest("release_datetime")
    original_duration = latest_ep.itunes_duration
    latest_ep.itunes_duration = 2460
    latest_ep.save()
    podcast_queryset = Podcast.objects.filter(id=podcast_with_parsed_episodes.id)
    request = rf.get("/admin/podcast_analyzer/")
    request.user = user
    request._messages = messages.storage.default_storage(request)
    admin_obj = PodcastAdmin(Podcast, admin.site)
    eps, feeds = admin_obj.feed_update(
        request, podcast_queryset, update_existing_episodes=True
    )
    latest_ep.refresh_from_db()
    assert latest_ep.itunes_duration == original_duration
    assert eps == 5
    assert feeds == 1


def test_feed_unreachable(
    rf, settings, httpx_mock, user, podcast_with_parsed_episodes
) -> None:
    settings.MESSAGE_STORAGE = "django.contrib.messages.storage.cookie.CookieStorage"
    httpx_mock.add_response(url=podcast_with_parsed_episodes.rss_feed, status_code=404)
    podcast_queryset = Podcast.objects.filter(id=podcast_with_parsed_episodes.id)
    request = rf.get("/admin/podcast_analyzer/")
    request.user = user
    request._messages = messages.storage.default_storage(request)
    admin_obj = PodcastAdmin(Podcast, admin.site)
    eps, feeds = admin_obj.feed_update(
        request, podcast_queryset, update_existing_episodes=True
    )
    assert eps == 0
    assert feeds == 1


def test_feed_invalid(
    rf, settings, httpx_mock, user, podcast_with_parsed_episodes
) -> None:
    settings.MESSAGE_STORAGE = "django.contrib.messages.storage.cookie.CookieStorage"
    with open("tests/data/malformed_podcast_rss_feed.xml", "rb") as f:
        datastream = BytesIO(f.read())
    httpx_mock.add_response(
        url=podcast_with_parsed_episodes.rss_feed, content=datastream
    )
    podcast_queryset = Podcast.objects.filter(id=podcast_with_parsed_episodes.id)
    request = rf.get("/admin/podcast_analyzer/")
    request.user = user
    request._messages = messages.storage.default_storage(request)
    admin_obj = PodcastAdmin(Podcast, admin.site)
    eps, feeds = admin_obj.feed_update(
        request, podcast_queryset, update_existing_episodes=True
    )
    assert eps == 0
    assert feeds == 1



# settings.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""Django settings to use for local development and test suite."""

from pathlib import Path

ROOT_DIR = Path(__file__).resolve(strict=True).parent.parent
APPS_DIR = Path(ROOT_DIR, "src")
ADMIN_URL = "admin/"

DEBUG = True

TIMEZONE = "UTC"
LANGUAGE_CODE = "en-us"
SITE_ID = 1
USE_I18N = True
USE_TZ = True

ALLOWED_HOSTS = ["*"]


LOCALE_PATHS = [
    str(APPS_DIR / "podcast_analyzer/locale"),
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "podcast_analyzer.sqlite",
        "ATOMIC_REQUESTS": True,
    }
}

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

ROOT_URLCONF = "tests.urls"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    # "django.contrib.humanize", # Handy template tags
    "django.contrib.admin",
    "django.forms",
    "tagulous",
    "django_browser_reload",
    "django_q",
    "django_watchfiles",
    "django_extensions",
    "podcast_analyzer",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "debug_toolbar.middleware.DebugToolbarMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django_browser_reload.middleware.BrowserReloadMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    # "django.middleware.common.BrokenLinkEmailsMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

# MEDIA
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#media-root
MEDIA_ROOT = str(ROOT_DIR / "media")
# https://docs.djangoproject.com/en/dev/ref/settings/#media-url
MEDIA_URL = "/media/"

STATIC_ROOT = str(ROOT_DIR / "staticfiles")
STATIC_URL = "/static/"
STATICFILES_DIRS = [str(APPS_DIR / "podcast_analyzer/static")]

# TEMPLATES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#templates
TEMPLATES = [
    {
        # https://docs.djangoproject.com/en/dev/ref/settings/#std:setting-TEMPLATES-BACKEND
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        # https://docs.djangoproject.com/en/dev/ref/settings/#dirs
        "DIRS": [str(APPS_DIR / "templates")],
        # https://docs.djangoproject.com/en/dev/ref/settings/#app-dirs
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": DEBUG,
            # https://docs.djangoproject.com/en/dev/ref/settings/#template-context-processors
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    }
]

# EMAIL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#email-backend
EMAIL_BACKEND = "django.core.mail.backends.locmem.EmailBackend"

SECRET_KEY = "django-insecure-jDtBEAbvuJosEg1frjvmNX4HmA46"  # Doesn't matter since this is a test.

# https://docs.djangoproject.com/en/dev/ref/settings/#test-runner
TEST_RUNNER = "django.test.runner.DiscoverRunner"

SERIALIZATION_MODULES = {
    "xml": "tagulous.serializers.xml_serializer",
    "json": "tagulous.serializers.json",
    "python": "tagulous.serializers.python",
    "yaml": "tagulous.serializers.pyyaml",
}

Q_CLUSTER = {
    "name": "podcast-analyzer",
    "compress": True,
    "timeout": 120,
    "retry": 240,
    "save_limit": 250,
    "label": "Django Q2",
    "orm": "default",
    "sync": False,
}


CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
    }
}


# LOGGING
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#logging
# See https://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": "%(levelname)s %(asctime)s %(module)s "
            "%(process)d %(thread)d %(message)s"
        }
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        }
    },
    "root": {"level": "INFO", "handlers": ["console"]},
    "loggers": {
        "podcast_analyzer": {
            "handlers": ["console"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
}



# Generated by Django 5.1.2 on 2024-11-21 14:16

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("podcast_analyzer", "0005_analysisgroup_description"),
    ]

    operations = [
        migrations.AddField(
            model_name="person",
            name="merged_at",
            field=models.DateTimeField(
                blank=True, help_text="When the record was merged", null=True
            ),
        ),
        migrations.AddField(
            model_name="person",
            name="merged_into",
            field=models.ForeignKey(
                blank=True,
                help_text="A primary record this person has been merged into.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="merged_records",
                to="podcast_analyzer.person",
            ),
        ),
    ]



# forms.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from django import forms
from django.core.exceptions import ValidationError

from podcast_analyzer.models import AnalysisGroup, Episode, Person, Podcast, Season


class AnalysisGroupForm(forms.ModelForm):
    """
    Form class that enables setting the reverse relationships
    to Podcast, Season, and Episode.

    Attributes:
        name (forms.CharField): Name of the group
        description (forms.TextField): Description of the group
        podcasts (form.ModelMultipleChoiceField): List of podcasts
        seasons (form.ModelMultipleChoiceField): List of seasons
        episodes (form.ModelMultipleChoiceField): List of episodes
    """

    class Meta:
        model = AnalysisGroup
        fields = ["name", "description"]

    podcasts = forms.ModelMultipleChoiceField(
        queryset=Podcast.objects.all(), required=False
    )
    seasons = forms.ModelMultipleChoiceField(
        queryset=Season.objects.all(), required=False
    )
    episodes = forms.ModelMultipleChoiceField(
        queryset=Episode.objects.all(), required=False
    )


class PersonMergeForm(forms.Form):
    """
    Form class for validating person merge operations.

    Attributes:
        source_person (forms.ModelChoiceField): Source person
        destination_person (forms.ModelChoiceField): Destination person
    """

    cleaned_data = {}
    source_person = forms.ModelChoiceField(
        queryset=Person.objects.filter(merged_into__isnull=True),
        required=True,
        widget=forms.widgets.HiddenInput,
    )
    destination_person = forms.ModelChoiceField(
        queryset=Person.objects.filter(merged_into__isnull=True),
        required=True,
        widget=forms.widgets.HiddenInput,
    )

    def clean(self):
        """
        Run standard clean and then validate that user is not trying
        to merge a record into itself.
        """
        self.cleaned_data = super().clean()
        dest = self.cleaned_data.get("destination_person")
        source = self.cleaned_data.get("source_person")
        if dest == source:
            msg = "A record cannot be merged into itself!"
            raise ValidationError(msg)
        return self.cleaned_data



# urls.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from django.urls import path

from podcast_analyzer import views

app_name = "podcast_analyzer"

urlpatterns = [
    path("", view=views.AppEntryView.as_view(), name="entry"),
    path(
        "analysis-groups/", view=views.AnalysisGroupListView.as_view(), name="ag-list"
    ),
    path(
        "analysis-groups/create/",
        view=views.AnalysisGroupCreateView.as_view(),
        name="ag-create",
    ),
    path(
        "analysis-groups/<uuid:id>/",
        view=views.AnalysisGroupDetailView.as_view(),
        name="ag-detail",
    ),
    path(
        "analysis-groups/<uuid:id>/edit/",
        view=views.AnalysisGroupUpdateView.as_view(),
        name="ag-edit",
    ),
    path(
        "analysis-groups/<uuid:id>/delete/",
        view=views.AnalysisGroupDeleteView.as_view(),
        name="ag-delete",
    ),
    path("people/", view=views.PersonListView.as_view(), name="person-list"),
    path(
        "people/<uuid:id>/", view=views.PersonDetailView.as_view(), name="person-detail"
    ),
    path(
        "people/<uuid:id>/merge/",
        view=views.PersonMergeListView.as_view(),
        name="person-merge-list",
    ),
    path(
        "people/<uuid:id>/merge/<uuid:destination_id>/",
        view=views.PersonMergeView.as_view(),
        name="person-merge",
    ),
    path(
        "people/<uuid:id>/edit/",
        view=views.PersonUpdateView.as_view(),
        name="person-edit",
    ),
    path(
        "people/<uuid:id>/delete/",
        view=views.PersonDeleteView.as_view(),
        name="person-delete",
    ),
    path("podcasts/", view=views.PodcastListView.as_view(), name="podcast-list"),
    path(
        "podcasts/add/", view=views.PodcastCreateView.as_view(), name="podcast-create"
    ),
    path(
        "podcasts/<uuid:id>/",
        view=views.PodcastDetailView.as_view(),
        name="podcast-detail",
    ),
    path(
        "podcasts/<uuid:podcast_id>/episodes/",
        view=views.EpisodeListView.as_view(),
        name="episode-list",
    ),
    path(
        "podcasts/<uuid:podcast_id>/episodes/<uuid:id>/",
        view=views.EpisodeDetailView.as_view(),
        name="episode-detail",
    ),
    path(
        "podcasts/<uuid:podcast_id>/episodes/<uuid:id>/edit/",
        view=views.EpisodeUpdateView.as_view(),
        name="episode-edit",
    ),
    path(
        "podcasts/<uuid:podcast_id>/episodes/<uuid:id>/delete/",
        view=views.EpisodeDeleteView.as_view(),
        name="episode-delete",
    ),
    path(
        "podcasts/<uuid:id>/edit/",
        view=views.PodcastUpdateView.as_view(),
        name="podcast-edit",
    ),
    path(
        "podcasts/<uuid:id>/delete/",
        view=views.PodcastDeleteView.as_view(),
        name="podcast-delete",
    ),
    path(
        "tags/<slug:tag_slug>/",
        view=views.TagPodcastListView.as_view(),
        name="tag-podcast-list",
    ),
]



# Generated by Django 5.1.2 on 2024-10-17 19:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("podcast_analyzer", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="person",
            name="img_url",
            field=models.URLField(
                blank=True, help_text="URL of the person's avatar image.", null=True
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="url",
            field=models.URLField(
                blank=True, help_text="Website link for the person", null=True
            ),
        ),
    ]



# test_forms.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import pytest
from django.core.exceptions import ValidationError

from podcast_analyzer.forms import PersonMergeForm

pytestmark = pytest.mark.django_db


def test_disallow_merge_self(podcast_with_parsed_episodes):
    source_person = (
        podcast_with_parsed_episodes.episodes.first().hosts_detected_from_feed.first()
    )
    form = PersonMergeForm(
        initial={
            "source_person": source_person.id,
            "destination_person": source_person.id,
        }
    )
    with pytest.raises(ValidationError):
        form.clean()



# apps.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class PodcastAnalyzerConfig(AppConfig):
    name = "podcast_analyzer"
    verbose_name = _("Podcast Analyzer")
    default_auto_field = "django.db.models.AutoField"

    def ready(self) -> None:
        try:
            import podcast_analyzer.receivers  # noqa F401
        except ImportError:  # no cov
            pass



# __init__.py
#
# Copyright (c) 2024 - 2025 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""Django Podcast Analyzer"""

__version__ = "0.6.0"

from podcast_analyzer.exceptions import (
    FeedFetchError,
    FeedParseError,
    ImageRetrievalError,
)

__all__ = ["FeedFetchError", "FeedParseError", "ImageRetrievalError"]



# Generated by Django 5.1.2 on 2024-11-26 22:02

import podcast_analyzer.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("podcast_analyzer", "0006_person_merged_at_person_merged_into"),
    ]

    operations = [
        migrations.AddField(
            model_name="person",
            name="avatar",
            field=models.ImageField(
                blank=True,
                null=True,
                upload_to=podcast_analyzer.models.avatar_directory_path,
            ),
        ),
    ]



# Generated by Django 5.1.2 on 2024-10-21 20:49

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("podcast_analyzer", "0003_alter_podcast_itunes_categories"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="episode",
            name="tags",
        ),
        migrations.DeleteModel(
            name="Tagulous_Episode_tags",
        ),
    ]



# podcast.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from datetime import datetime, timedelta

import factory
import pytest
from django.utils import timezone
from factory.django import DjangoModelFactory
from factory.fuzzy import FuzzyChoice

from podcast_analyzer.models import (
    KNOWN_GENERATOR_HOST_MAPPING,
    Episode,
    Podcast,
    Season,
)

pytestmark = pytest.mark.django_db(transaction=True)


class PodcastFactory(DjangoModelFactory):
    """
    Generates a podcast object.
    """

    title = factory.Faker("text", max_nb_chars=50)
    rss_feed = factory.Faker("uri")
    podcast_cover_art_url = factory.LazyAttribute(
        lambda o: factory.Faker("uri") if o.cover_art else None
    )
    podcast_art_cache_update_needed = factory.LazyAttribute(
        lambda o: True if o.cover_art else False
    )
    author = factory.Faker("name")
    generator = factory.LazyAttribute(
        lambda o: None if o.empty else FuzzyChoice(KNOWN_GENERATOR_HOST_MAPPING.keys())
    )
    email = factory.LazyAttribute(lambda o: None if o.empty else factory.Faker("email"))
    site_url = factory.LazyAttribute(
        lambda o: None if o.empty else factory.Faker("url")
    )

    class Meta:
        model = Podcast

    class Params:
        empty = True
        cover_art = False


class EpisodeFactory(DjangoModelFactory):
    """
    Generates an episode.
    """

    podcast = factory.SubFactory(PodcastFactory)
    guid = factory.Faker("uri")
    title = factory.Faker("text", max_nb_chars=50)
    episode_url = factory.Faker("uri")
    download_url = factory.Faker("uri")
    release_datetime = timezone.now() - timedelta(days=365)
    ep_num = 1
    ep_type = "full"
    season = None

    class Meta:
        model = Episode


@pytest.mark.django_db
def generate_episodes_for_podcast(
    podcast: Podcast,
    number_of_episodes: int = 10,
    latest_datetime: datetime | None = None,
    tracking_data: bool = False,
    days_between: int = 7,
    add_bonus_episode: bool = False,
    season: Season | None = None,
) -> None:
    """
    Given a podcast object, generate episodes for it.

    Args:
        podcast (Podcast): The Podcast instance to create for.
        number_of_episodes (int): Number of episodes to create, defaults to 10.
        latest_datetime (datetime | None): Most recent datetime of last episode.
        tracking_data (bool): Whether there should be tracking data in the urls.
        days_between (int): How many days between episodes.
        add_bonus_episode (bool): Add an episode with type of bonus
        season (Season):The season to add the episodes to
    """
    if number_of_episodes <= 0:
        msg = "Episodes to create must be greater than 0!"
        raise ValueError(msg)
    if not latest_datetime:
        latest_datetime = timezone.now()
    base_url: str = str(factory.Faker("url", schemes=["https"]))
    if tracking_data:
        base_url = f"https://media.blubrry.com/accountname/{base_url[8:]}"
    current_datetime = (
        latest_datetime
        - timedelta(days=1)
        - timedelta(days=days_between * number_of_episodes)
    )
    for x in range(number_of_episodes):
        EpisodeFactory(
            podcast=podcast,
            episode_url=f"{base_url}/episodes/{x + 1}",
            download_url=f"{base_url}/media/{x + 1}.mp3",
            ep_num=x + 1,
            release_datetime=current_datetime,
            season=season,
        )
        current_datetime = current_datetime + timedelta(days=days_between)
    if add_bonus_episode:
        EpisodeFactory(
            podcast=podcast,
            episode_url=f"{base_url}/episodes/{number_of_episodes + 1}",
            download_url=f"{base_url}/media/{number_of_episodes +1 }.mp3",
            ep_num=number_of_episodes + 1,
            release_datetime=current_datetime + timedelta(days=1),
            season=season,
        )



# tasks.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
import logging

from podcast_analyzer.models import Person, Podcast

logger = logging.getLogger(__name__)


def async_refresh_feed(podcast_id: str) -> None:
    """
    Given a podcast object, call it's refresh feed function.

    Args:
        podcast_id (str): The podcast ID.
    """
    podcast = Podcast.objects.get(id=podcast_id)
    podcast.refresh_feed()


def run_feed_analysis(podcast: Podcast) -> None:
    """
    Wraps around the instance's feed analysis function.

    Args:
        podcast (Podcast): Podcast object to analyze.
    """
    logger.debug("Task 'run_feed_analysis' called!")
    asyncio.run(podcast.analyze_feed())
    podcast.schedule_next_refresh()


def fetch_podcast_cover_art(podcast: Podcast) -> None:
    """
    Wraps around the remote fetch functions for cover art.

    Args:
        podcast (Podcast): Podcast object to get cover art for.
    """
    logger.debug(f"Task 'fetch_podcast_cover_art' called for {podcast.title}!")
    asyncio.run(podcast.afetch_podcast_cover_art())


def fetch_avatar_for_person(person: Person) -> None:
    """
    Triggers the remote fetch of the person's avatar.

    Args:
        person (Person): The person record to populate with the avatar.
    """
    logger.debug(
        f"Task 'fetch_avatar_for_person' called for {person.name} ({person.id})"
    )
    asyncio.run(person.afetch_avatar())



# Generated by Django 5.1.2 on 2024-10-21 18:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("podcast_analyzer", "0002_alter_person_img_url_alter_person_url"),
    ]

    operations = [
        migrations.AlterField(
            model_name="podcast",
            name="itunes_categories",
            field=models.ManyToManyField(
                blank=True,
                related_name="podcasts",
                to="podcast_analyzer.itunescategory",
            ),
        ),
    ]



# test_receivers.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""Placeholder for receiver related tests."""

import pytest

from podcast_analyzer import receivers
from podcast_analyzer.models import Person, Podcast

pytestmark = pytest.mark.django_db(transaction=True)


def test_podcast_create_receiver(mocker):
    mocker.patch("podcast_analyzer.receivers.async_task")
    Podcast.objects.create(
        title="Some Techbros Chatting", rss_feed="https://example.com/podcast.rss"
    )
    receivers.async_task.assert_called_once()


def test_person_save_receiver_called(mocker):
    mocker.patch("podcast_analyzer.receivers.async_task")
    Person.objects.create(
        name="John Doe",
        url="https://example.com",
        img_url="https://example.com/people/jdoe/avatar.png",
    )
    receivers.async_task.assert_called_once()



# exceptions.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

class FeedFetchError(Exception):
    """
    Raised when we are unable to retrieve remote feed data due to either
    server error or network connectivity.
    """

    pass


class FeedParseError(Exception):
    """
    Raised when parsing the content of a feed response results
    in a parsing error.
    """

    pass


class ImageRetrievalError(Exception):
    """
    Raised when we are unable to retrieve image data from a remote source.
    """

    pass



# test_image_fetching.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from io import BytesIO

import pytest
from django.core.files import File

from podcast_analyzer import ImageRetrievalError
from podcast_analyzer.models import RemoteImageData, fetch_image_for_record


@pytest.mark.parametrize(
    "url,reported_mime_type,expected_filename",
    [
        ("https://example.com/imgs/cover.jpg", "image/jpeg", "cover.png"),
        (
            "https://example.com/imgs/podcasts/20/cover_art.jpg",
            "image/jpg",
            "cover_art.png",
        ),
        (
            "https://example.com/imgs/cover.jpg?size=200,200",
            "application/octet-stream",
            "cover.png",
        ),
    ],
)
def test_image_data_initialization(
    cover_art, url, reported_mime_type, expected_filename
):
    img_bytes = BytesIO(cover_art)
    rid = RemoteImageData(
        img_data=img_bytes, remote_url=url, reported_mime_type="image/jpeg"
    )
    assert rid.actual_mime_type == "image/png"
    assert rid.image_file is not None and isinstance(rid.image_file, File)
    assert rid.filename == expected_filename
    assert rid.is_valid()


@pytest.mark.parametrize(
    "allowed_mime_types,expect_valid",
    [
        (["image/jpeg", "image/gif", "image/webp"], False),
        (["image/jpeg", "image/gif", "image/png", "image/png"], True),
    ],
)
def test_image_data_initialization_with_different_mime_type(
    cover_art, allowed_mime_types, expect_valid
):
    img_bytes = BytesIO(cover_art)
    url = "https://example.com/imgs/podcasts/20/cover_art.png"
    rid = RemoteImageData(
        img_data=img_bytes,
        remote_url=url,
        reported_mime_type="image/png",
        allowed_mime_types=allowed_mime_types,
    )
    assert rid._allowed_mime_types == allowed_mime_types
    assert expect_valid == rid.is_valid()


@pytest.mark.django_db
def test_image_data_initialization_with_invalid(rss_feed_datastream):
    rid = RemoteImageData(
        img_data=rss_feed_datastream,
        remote_url="https://example.com/cover.jpg",
        reported_mime_type="image/jpeg",
    )
    assert not rid.is_valid()


@pytest.mark.asyncio
@pytest.mark.parametrize("status_code", [400, 401, 403, 404, 500])
async def test_fetch_image_for_record_http_error(httpx_mock, status_code):
    url = "https://example.com/imgs/cover.png"
    httpx_mock.add_response(url=url, status_code=status_code)
    with pytest.raises(ImageRetrievalError):
        await fetch_image_for_record(img_url=url)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url,reported_mime_type,expected_filename",
    [
        ("https://example.com/imgs/cover.jpg", "image/jpeg", "cover.png"),
        (
            "https://example.com/imgs/cover.jpg?size=200,200",
            "application/octet-stream",
            "cover.png",
        ),
        ("https://example.com/imgs/cover.png", "image/png", "cover.png"),
    ],
)
async def test_fetch_image_for_record_success(
    httpx_mock, cover_art, url, reported_mime_type, expected_filename
):
    headers = {"Content-Type": reported_mime_type}
    httpx_mock.add_response(url=url, headers=headers, content=cover_art)
    rid = await fetch_image_for_record(img_url=url)
    assert rid.filename == expected_filename
    assert rid.actual_mime_type == "image/png"
    assert rid.image_file is not None and isinstance(rid.image_file, File)



# test_models.py
#
# Copyright (c) 2024 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import os
from datetime import datetime, timedelta
from io import BytesIO
from statistics import median_high

import pytest
from asgiref.sync import sync_to_async
from django.utils import timezone
from django_q.models import Schedule

from podcast_analyzer.exceptions import FeedFetchError, FeedParseError
from podcast_analyzer.models import (
    Episode,
    ItunesCategory,
    Person,
    Podcast,
    calculate_median_episode_duration,
)
from tests.factories.podcast import PodcastFactory, generate_episodes_for_podcast

pytestmark = pytest.mark.django_db(transaction=True)


def test_get_missing_feed(httpx_mock, empty_podcast):
    httpx_mock.add_response(url=empty_podcast.rss_feed, status_code=404)
    with pytest.raises(FeedFetchError):
        empty_podcast.get_feed_data()


def test_get_malformed_feed(httpx_mock, empty_podcast):
    with open(
        "tests/data/malformed_podcast_rss_feed.xml",
        "rb",
    ) as f:
        malformed_bytes = f.read()
    httpx_mock.add_response(url=empty_podcast.rss_feed, content=malformed_bytes)
    with pytest.raises(FeedParseError):
        empty_podcast.get_feed_data()


def test_get_feed_data(httpx_mock, empty_podcast, rss_feed_datastream):
    httpx_mock.add_response(url=empty_podcast.rss_feed, content=rss_feed_datastream)
    result = empty_podcast.get_feed_data()
    assert result["title"] == "Some Podcast"
    assert len(result["episodes"]) == 5


def test_update_metadata(httpx_mock, empty_podcast, rss_feed_datastream):
    httpx_mock.add_response(url=empty_podcast.rss_feed, content=rss_feed_datastream)
    result = empty_podcast.get_feed_data()
    empty_podcast.update_podcast_metadata_from_feed_data(result)
    assert empty_podcast.feed_contains_itunes_data
    assert empty_podcast.itunes_explicit
    assert empty_podcast.feed_contains_podcast_index_data
    assert empty_podcast.feed_contains_structured_donation_data
    assert empty_podcast.itunes_categories.count() == 2
    assert empty_podcast.podcast_cover_art_url is not None
    assert empty_podcast.podcast_art_cache_update_needed
    assert empty_podcast.author == "Some Podcast Company"
    assert empty_podcast.email == "contact@somepodcast.com"
    assert empty_podcast.last_checked is not None


def test_update_metadata_no_itunes_owner(empty_podcast):
    feed_dict = {
        "title": "Some Podcast",
        "episodes": [],
        "link": "https://www.somepodcast.com",
        "generator": "https://podbean.com/?v=5.5",
        "language": "en",
        "type": "serial",
        "itunes_author": "Some Podcast Company",
        "cover_url": "https://media.somepodcast.com/cover.jpg",
        "explicit": True,
        "itunes_keywords": ["games", "fiction", "unbearable opinions"],
        "import_prohibited": True,
        "funding_url": "https://www.patreonclone.com/somepodcast",
        "itunes_categories": [["Leisure", "Games"], ["Fiction", "Comedy Fiction"]],
    }
    empty_podcast.update_podcast_metadata_from_feed_data(feed_dict)
    assert empty_podcast.email is None


def test_update_metadata_category_no_children(empty_podcast):
    feed_dict = {
        "title": "Some Podcast",
        "episodes": [],
        "link": "https://www.somepodcast.com",
        "generator": "https://podbean.com/?v=5.5",
        "language": "en",
        "type": "serial",
        "itunes_author": "Some Podcast Company",
        "cover_url": "https://media.somepodcast.com/cover.jpg",
        "explicit": True,
        "itunes_keywords": ["games", "fiction", "unbearable opinions"],
        "import_prohibited": True,
        "funding_url": "https://www.patreonclone.com/somepodcast",
        "itunes_categories": [["Leisure", "Games"], ["Fiction"]],
    }
    empty_podcast.update_podcast_metadata_from_feed_data(feed_dict)
    fic_cat = ItunesCategory.objects.get(name="Fiction")
    assert fic_cat in empty_podcast.itunes_categories.all()


def test_update_metadata_no_podcast_index(empty_podcast):
    feed_dict = {
        "title": "Some Podcast",
        "episodes": [],
        "link": "https://www.somepodcast.com",
        "generator": "https://podbean.com/?v=5.5",
        "language": "en",
        "type": "serial",
        "itunes_author": "Some Podcast Company",
        "cover_url": "https://media.somepodcast.com/cover.jpg",
        "explicit": True,
        "itunes_keywords": ["games", "fiction", "unbearable opinions"],
        "itunes_owner": {
            "name": "Some Podcast Company",
            "email": "contact@somepodcast.com",
        },
        "itunes_categories": [["Leisure", "Games"], ["Fiction", "Comedy Fiction"]],
    }
    empty_podcast.update_podcast_metadata_from_feed_data(feed_dict)
    assert not empty_podcast.feed_contains_podcast_index_data


def test_update_metadata_no_itunes(empty_podcast):
    feed_dict = {
        "title": "Some Podcast",
        "episodes": [],
        "link": "https://www.somepodcast.com",
        "generator": "https://podbean.com/?v=5.5",
        "language": "en",
        "type": "serial",
        "cover_url": "https://media.somepodcast.com/cover.jpg",
        "explicit": True,
        "import_prohibited": True,
        "funding_url": "https://www.patreonclone.com/somepodcast",
    }
    empty_podcast.update_podcast_metadata_from_feed_data(feed_dict)
    assert not empty_podcast.feed_contains_itunes_data


def test_update_metadata_no_donation(empty_podcast):
    feed_dict = {
        "title": "Some Podcast",
        "episodes": [],
        "link": "https://www.somepodcast.com",
        "generator": "https://podbean.com/?v=5.5",
        "language": "en",
        "type": "serial",
        "itunes_author": "Some Podcast Company",
        "cover_url": "https://media.somepodcast.com/cover.jpg",
        "explicit": True,
        "itunes_keywords": ["games", "fiction", "unbearable opinions"],
        "itunes_owner": {
            "name": "Some Podcast Company",
            "email": "contact@somepodcast.com",
        },
        "import_prohibited": True,
        "itunes_categories": [["Leisure", "Games"], ["Fiction", "Comedy Fiction"]],
    }
    empty_podcast.update_podcast_metadata_from_feed_data(feed_dict)
    assert not empty_podcast.feed_contains_structured_donation_data


def test_update_metadata_no_cover_art(empty_podcast):
    feed_dict = {
        "title": "Some Podcast",
        "episodes": [],
        "link": "https://www.somepodcast.com",
        "generator": "https://podbean.com/?v=5.5",
        "language": "en",
        "type": "serial",
        "itunes_author": "Some Podcast Company",
        "explicit": True,
        "itunes_keywords": ["games", "fiction", "unbearable opinions"],
        "itunes_owner": {
            "name": "Some Podcast Company",
            "email": "contact@somepodcast.com",
        },
        "import_prohibited": True,
        "funding_url": "https://www.patreonclone.com/somepodcast",
        "itunes_categories": [["Leisure", "Games"], ["Fiction", "Comedy Fiction"]],
    }
    empty_podcast.update_podcast_metadata_from_feed_data(feed_dict)
    assert empty_podcast.podcast_cover_art_url is None


@pytest.mark.parametrize(
    "cover_url,response_status,response_headers,expect_success",
    [
        (None, None, None, False),
        ("https://media.somepodcast.com/cover.jpg", 404, None, False),
        ("https://media.somepodcast.com/cover.jpg", 200, None, False),
        (
            "https://media.somepodcast.com/cover.jpg",
            200,
            [("Content-Type", "image/jpeg")],
            True,
        ),
    ],
)
def test_fetch_cover_art(
    httpx_mock,
    valid_podcast,
    cover_url,
    response_status,
    response_headers,
    cover_art,
    expect_success,
):
    if cover_url is not None:
        valid_podcast.podcast_cover_art_url = cover_url
        valid_podcast.podcast_art_cache_update_needed = True
        valid_podcast.save()
        if response_status == 200 and response_headers is not None:
            httpx_mock.add_response(
                url=cover_url, headers=response_headers, content=cover_art
            )
        else:
            httpx_mock.add_response(url=cover_url, status_code=response_status)
        valid_podcast.fetch_podcast_cover_art()
        if expect_success:
            assert not valid_podcast.podcast_art_cache_update_needed
            assert valid_podcast.podcast_cached_cover_art
        else:
            assert valid_podcast.podcast_art_cache_update_needed
            assert not valid_podcast.podcast_cached_cover_art
    else:
        assert valid_podcast.podcast_cover_art_url is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cover_url,response_status,response_headers,expect_success",
    [
        (None, None, None, False),
        ("https://media.somepodcast.com/cover.jpg", 404, None, False),
        ("https://media.somepodcast.com/cover.jpg", 200, None, False),
        (
            "https://media.somepodcast.com/cover.jpg",
            200,
            [("Content-Type", "image/jpeg")],
            True,
        ),
        (
            "https://media.somepodcast.com/cover.jpg?from=rss",
            200,
            [("Content-Type", "image/jpeg")],
            True,
        ),
    ],
)
async def test_async_fetch_cover_art(
    httpx_mock,
    valid_podcast,
    cover_url,
    response_status,
    response_headers,
    cover_art,
    expect_success,
):
    if cover_url is not None:
        valid_podcast.podcast_cover_art_url = cover_url
        valid_podcast.podcast_art_cache_update_needed = True
        await valid_podcast.asave()
        if response_status == 200 and response_headers is not None:
            httpx_mock.add_response(
                url=cover_url, headers=response_headers, content=cover_art
            )
        else:
            httpx_mock.add_response(url=cover_url, status_code=response_status)
        await valid_podcast.afetch_podcast_cover_art()
        if expect_success:
            assert not valid_podcast.podcast_art_cache_update_needed
            assert valid_podcast.podcast_cached_cover_art
            assert valid_podcast.podcast_cached_cover_art.name[-3:] == "png"
        else:
            assert valid_podcast.podcast_art_cache_update_needed
            assert not valid_podcast.podcast_cached_cover_art
    else:
        assert valid_podcast.podcast_cover_art_url is None


def test_sync_fetch_cover_art_invalid_file(httpx_mock, valid_podcast):
    file_size = 9067
    random_file = BytesIO(initial_bytes=os.urandom(file_size))
    cover_url = "https://media.somepodcast.com/cover.jpg?from=rss"
    httpx_mock.add_response(
        url=cover_url,
        headers=[("Content-Type", "image/jpeg")],
        content=random_file.read(),
    )
    valid_podcast.podcast_cover_art_url = cover_url
    valid_podcast.podcast_art_cache_update_needed = True
    valid_podcast.save()
    valid_podcast.fetch_podcast_cover_art()
    art_update = valid_podcast.art_updates.latest("timestamp")
    assert not art_update.valid_file


@pytest.mark.parametrize(
    "update_all_eps,expected_first_touch_count,expected_second_touch_count",
    [
        (False, 5, 0),
        (True, 5, 5),
    ],
)
def test_new_episodes_in_feed(
    empty_podcast,
    parsed_rss,
    update_all_eps,
    expected_first_touch_count,
    expected_second_touch_count,
):
    empty_podcast.update_podcast_metadata_from_feed_data(parsed_rss)
    first_touch = empty_podcast.update_episodes_from_feed_data(
        parsed_rss["episodes"], update_existing_episodes=update_all_eps
    )
    assert first_touch == expected_first_touch_count
    second_touch = empty_podcast.update_episodes_from_feed_data(
        parsed_rss["episodes"], update_existing_episodes=update_all_eps
    )
    assert second_touch == expected_second_touch_count


def test_episodes_contain_funding_data(empty_podcast, parsed_rss):
    assert not empty_podcast.feed_contains_structured_donation_data
    parsed_rss["episodes"][0]["payment_url"] = "https://ko-fi.com/somepodcast"
    empty_podcast.update_episodes_from_feed_data(
        episode_list=parsed_rss["episodes"], update_existing_episodes=True
    )
    empty_podcast.refresh_from_db()
    assert empty_podcast.feed_contains_structured_donation_data


@pytest.mark.asyncio
async def test_analyze_host(podcast_with_parsed_metadata):
    await podcast_with_parsed_metadata.analyze_host()
    assert podcast_with_parsed_metadata.probable_feed_host == "Podbean"


@pytest.mark.asyncio
async def test_analyze_host_known_generator(mute_signals):
    new_podcast = await Podcast.objects.acreate(
        title="Tech Bros BSing",
        rss_feed="https://example.com",
        generator="Fireside (https://fireside.fm)",
    )
    await new_podcast.analyze_host()
    assert new_podcast.probable_feed_host == "Fireside.fm"
    await new_podcast.adelete()


@pytest.mark.asyncio
async def test_analyze_empty_host(mute_signals):
    new_podcast = await Podcast.objects.acreate(
        title="Tech Bros BSing", rss_feed="https://example.com", generator="monkey"
    )
    await new_podcast.analyze_host()
    assert new_podcast.probable_feed_host is None
    await new_podcast.adelete()


@pytest.mark.asyncio
async def test_analyze_host_from_episodes(mute_signals):
    new_podcast = await Podcast.objects.acreate(
        title="Tech Bros BSing", rss_feed="https://example.com", generator="monkey"
    )
    for i in range(5):
        await Episode.objects.acreate(
            podcast=new_podcast,
            guid=new_podcast.title.replace(" ", "-") + f"--{i}",
            download_url=f"http://media.blubrry.com/somepodcast/{i}",
        )
    await new_podcast.analyze_host()
    assert new_podcast.probable_feed_host == "Blubrry"
    await new_podcast.adelete()


@pytest.mark.asyncio
async def test_find_tracking(active_tracking_podcast):
    await active_tracking_podcast.analyze_feed_for_third_party_analytics()
    assert active_tracking_podcast.feed_contains_tracking_data


@pytest.mark.asyncio
async def test_find_tracking_false(active_podcast):
    await active_podcast.analyze_feed_for_third_party_analytics()
    assert not active_podcast.feed_contains_tracking_data


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "days_between,expected_result",
    [
        (1, "daily"),
        (3, "often"),
        (7, "weekly"),
        (14, "biweekly"),
        (30, "monthly"),
        (45, "adhoc"),
    ],
)
async def test_calculate_release_frequency(
    podcast_with_parsed_metadata, days_between, expected_result
):
    await sync_to_async(generate_episodes_for_podcast)(
        podcast=podcast_with_parsed_metadata,
        latest_datetime=timezone.now() - timedelta(days=6),
        days_between=days_between,
    )

    await podcast_with_parsed_metadata.set_release_frequency(
        podcast_with_parsed_metadata.episodes.all()
    )
    assert podcast_with_parsed_metadata.release_frequency == expected_result


def test_duration_calculations(podcast_with_parsed_episodes):
    assert podcast_with_parsed_episodes.total_duration_seconds == 15702
    assert podcast_with_parsed_episodes.total_duration_timedelta == timedelta(
        seconds=15702
    )


def test_duration_empty_podcast(empty_podcast):
    assert empty_podcast.total_duration_seconds == 0
    assert empty_podcast.total_duration_timedelta is None


def test_ep_count_active(active_podcast):
    assert active_podcast.total_episodes == 10


def test_median_duration_active(podcast_with_parsed_episodes):
    ep_durations = [
        e.itunes_duration for e in podcast_with_parsed_episodes.episodes.all()
    ]
    assert (
        median_high(ep_durations)
        == podcast_with_parsed_episodes.median_episode_duration
    )
    assert podcast_with_parsed_episodes.median_episode_duration_timedelta == timedelta(
        seconds=podcast_with_parsed_episodes.median_episode_duration
    )


def test_median_duration_empty(empty_podcast):
    assert empty_podcast.median_episode_duration == 0
    assert empty_podcast.median_episode_duration_timedelta == timedelta(seconds=0)


def test_sync_last_release(podcast_with_parsed_episodes):
    expected_release_datetime = datetime.strptime(
        "Fri, 29 Apr 2023 06:00:00 -0400", "%a, %d %b %Y %H:%M:%S %z"
    )
    assert podcast_with_parsed_episodes.last_release_date == expected_release_datetime


def test_last_release_empty(empty_podcast):
    assert empty_podcast.last_release_date is None


@pytest.mark.asyncio
async def test_async_last_release(podcast_with_parsed_episodes):
    expected_release_datetime = datetime.strptime(
        "Fri, 29 Apr 2023 06:00:00 -0400", "%a, %d %b %Y %H:%M:%S %z"
    )
    rel_date = await podcast_with_parsed_episodes.alast_release_date()
    assert rel_date == expected_release_datetime


@pytest.mark.asyncio
async def test_async_last_release_empty(empty_podcast):
    assert await empty_podcast.alast_release_date() is None


def test_schedule_next_refresh_empty(empty_podcast):
    current_schedule_count = Schedule.objects.count()
    empty_podcast.schedule_next_refresh(last_release_date=None)
    assert current_schedule_count == Schedule.objects.count()


@pytest.mark.asyncio
async def test_detect_dormant(dormant_podcast):
    await dormant_podcast.set_dormant()
    assert dormant_podcast.dormant


@pytest.mark.asyncio
async def test_detect_dormant_empty(empty_podcast):
    last_mod = empty_podcast.modified
    await empty_podcast.set_dormant()
    await empty_podcast.arefresh_from_db()
    assert last_mod == empty_podcast.modified


@pytest.mark.asyncio
async def test_detect_active(active_podcast):
    await active_podcast.set_dormant()
    assert not active_podcast.dormant


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "full_episodes_only,latest_release,use_tracking,days_between,episode_limit,"
    "expected_freq,expected_dormant,expected_tracking",
    [
        (True, timezone.now() - timedelta(days=70), False, 1, 0, "daily", True, False),
        (True, timezone.now() - timedelta(days=70), True, 1, 0, "daily", True, True),
        (True, timezone.now() - timedelta(days=2), False, 1, 0, "daily", False, False),
        (False, timezone.now() - timedelta(days=2), False, 1, 0, "daily", False, False),
        (True, timezone.now() - timedelta(days=2), False, 3, 0, "often", False, False),
        (True, timezone.now() - timedelta(days=2), False, 7, 0, "weekly", False, False),
        (
            True,
            timezone.now() - timedelta(days=2),
            False,
            7,
            2,
            "pending",
            False,
            False,
        ),
        (
            True,
            timezone.now() - timedelta(days=2),
            False,
            14,
            0,
            "biweekly",
            False,
            False,
        ),
        (
            True,
            timezone.now() - timedelta(days=2),
            False,
            30,
            0,
            "monthly",
            False,
            False,
        ),
        (
            True,
            timezone.now() - timedelta(days=2),
            False,
            45,
            0,
            "adhoc",
            False,
            False,
        ),
    ],
)
async def test_analyze_podcast(
    podcast_with_parsed_metadata,
    episode_limit,
    full_episodes_only,
    latest_release,
    use_tracking,
    days_between,
    expected_freq,
    expected_dormant,
    expected_tracking,
):
    await sync_to_async(generate_episodes_for_podcast)(
        podcast=podcast_with_parsed_metadata,
        latest_datetime=latest_release,
        days_between=days_between,
        tracking_data=use_tracking,
        add_bonus_episode=True,
    )
    await podcast_with_parsed_metadata.analyze_feed(
        episode_limit=episode_limit, full_episodes_only=full_episodes_only
    )
    assert podcast_with_parsed_metadata.release_frequency == expected_freq
    assert podcast_with_parsed_metadata.dormant == expected_dormant
    assert podcast_with_parsed_metadata.feed_contains_tracking_data == expected_tracking


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "active,days_between,expected_delta",
    [
        (True, 1, timedelta(days=1)),
        (True, 4, timedelta(days=3)),
        (True, 7, timedelta(days=7)),
        (True, 14, timedelta(days=14)),
        (True, 29, timedelta(days=30)),
        (True, 45, timedelta(days=60)),
        (False, 7, timedelta(days=60)),
    ],
)
async def test_calculate_next_refresh(
    podcast_with_parsed_metadata, active, days_between, expected_delta
):
    if active:
        last_release = timezone.now()
    else:
        last_release = timezone.now() - timedelta(days=70)
    await sync_to_async(generate_episodes_for_podcast)(
        podcast=podcast_with_parsed_metadata,
        latest_datetime=last_release,
        days_between=days_between,
    )
    await podcast_with_parsed_metadata.analyze_feed()
    calculated_refresh = podcast_with_parsed_metadata.calculate_next_refresh_time(
        last_release_date=last_release
    )
    calculated_diff = calculated_refresh - last_release
    if active:
        assert calculated_diff == expected_delta
    else:
        assert calculated_diff >= expected_delta


def test_episode_persons_detection(podcast_with_parsed_episodes):
    """
    Checks that hosts and guests are correctly parsed from feed and that existing
    records are not duplicated.
    """
    episode = podcast_with_parsed_episodes.episodes.latest("release_datetime")
    assert episode.hosts_detected_from_feed.count() == 1
    assert episode.guests_detected_from_feed.count() == 1
    assert Person.objects.count() == 2


def test_episode_no_person_records(active_podcast):
    """
    Checks that we don't create erroneous records when no
    person elements appear in the episode feed.
    """
    episode = active_podcast.episodes.latest("release_datetime")
    assert episode.hosts_detected_from_feed.count() == 0
    assert episode.guests_detected_from_feed.count() == 0
    assert Person.objects.count() == 0


def test_detect_person_img(podcast_with_parsed_episodes):
    """
    Checks if the system has correctly set the img element when supplied
    and left null if not.
    """
    episode = podcast_with_parsed_episodes.episodes.latest("release_datetime")
    host = episode.hosts_detected_from_feed.all()[0]
    guest = episode.guests_detected_from_feed.all()[0]
    assert guest.img_url is None
    assert host.img_url is not None


@pytest.mark.parametrize(
    "podcast_count,skip_podcasts,hosted_eps,guested_eps,expected_result",
    [
        (2, 0, 4, 3, 2),
        (2, 1, 4, 0, 1),
        (2, 1, 3, 0, 1),
        (3, 1, 5, 2, 2),
        (3, 2, 0, 1, 1),
        (3, 3, 0, 0, 0),
    ],
)
def test_person_distinct_podcasts(
    mute_signals, podcast_count, skip_podcasts, hosted_eps, guested_eps, expected_result
):
    podcasts = [PodcastFactory() for _ in range(podcast_count)]
    for podcast in podcasts:
        generate_episodes_for_podcast(podcast)
    person = Person.objects.create(name="John Smith")
    if skip_podcasts < podcast_count:
        hosting_done = False
        for podcast in podcasts[skip_podcasts:]:
            if guested_eps > 0 and hosting_done or hosted_eps == 0:
                eps_to_edit = podcast.episodes.all()[:guested_eps]
                for ep in eps_to_edit:
                    ep.guests_detected_from_feed.add(person)
            elif hosted_eps > 0 and not hosting_done:
                eps_to_edit = podcast.episodes.all()[:hosted_eps]
                for ep in eps_to_edit:
                    ep.hosts_detected_from_feed.add(person)
                hosting_done = True
            else:
                pass  # No person entries to add.
    assert person.has_guested == guested_eps
    assert person.has_hosted == hosted_eps
    assert person.distinct_podcasts == expected_result


@pytest.mark.parametrize(
    "podcast_count,hosted_eps,guested_eps", [(0, 0, 0), (2, 4, 0), (2, 5, 5), (1, 0, 5)]
)
def test_person_total_episodes(mute_signals, podcast_count, hosted_eps, guested_eps):
    podcasts = [PodcastFactory() for _ in range(podcast_count)]
    for podcast in podcasts:
        generate_episodes_for_podcast(podcast)
    person = Person.objects.create(name="John Smith", url="https://www.example.com")
    hosting_done = False
    for podcast in podcasts:
        if guested_eps > 0 and hosting_done or hosted_eps == 0:
            eps_to_edit = podcast.episodes.all()[:guested_eps]
            for ep in eps_to_edit:
                ep.guests_detected_from_feed.add(person)
        elif hosted_eps > 0 and not hosting_done:
            eps_to_edit = podcast.episodes.all()[:hosted_eps]
            for ep in eps_to_edit:
                ep.hosts_detected_from_feed.add(person)
            hosting_done = True
        else:
            pass  # No entries to add
    assert person.get_total_episodes() == guested_eps + hosted_eps


def test_person_podcast_appearance_data(mute_signals):
    generated_podcasts = [PodcastFactory().id for _ in range(3)]
    person = Person.objects.create(name="John Smith", url="https://www.example.com")
    podcasts = Podcast.objects.filter(pk__in=generated_podcasts).order_by("title")
    for podcast in podcasts:
        generate_episodes_for_podcast(podcast)
    for ep in podcasts[0].episodes.all():
        ep.hosts_detected_from_feed.add(person)
    for ep in podcasts[1].episodes.all()[:5]:
        ep.guests_detected_from_feed.add(person)
    pod3_eps = podcasts[2].episodes.all()
    for ep in pod3_eps[:2]:
        ep.hosts_detected_from_feed.add(person)
    for ep in pod3_eps[4:6]:
        ep.guests_detected_from_feed.add(person)
    pod_data = person.get_podcasts_with_appearance_counts()
    assert len(pod_data) == 3
    assert pod_data[0].hosted_episodes.count() == 10
    assert pod_data[0].guested_episodes.count() == 0
    assert pod_data[1].hosted_episodes.count() == 0
    assert pod_data[1].guested_episodes.count() == 5
    assert pod_data[2].hosted_episodes.count() == 2
    assert pod_data[2].guested_episodes.count() == 2


@pytest.mark.asyncio
async def test_person_fetch_avatar_no_url():
    person = await Person.objects.acreate(
        name="John Doe", url="https://www.example.com"
    )
    await person.afetch_avatar()
    assert not person.avatar


@pytest.mark.asyncio
async def test_person_fetch_avatar_request_fail(mute_signals, httpx_mock):
    person = await Person.objects.acreate(
        name="John Doe",
        url="https://www.example.com",
        img_url="https://example.com/jdoe.jpg",
    )
    httpx_mock.add_response(url="https://example.com/jdoe.jpg", status_code=404)
    await person.afetch_avatar()
    assert not person.avatar


@pytest.mark.asyncio
async def test_person_fetch_avatar(mute_signals, cover_art, httpx_mock):
    person = await Person.objects.acreate(
        name="John Doe",
        url="https://www.example.com",
        img_url="https://example.com/jdoe.jpg",
    )
    headers = {"Content-Type": "image/png"}
    httpx_mock.add_response(
        url="https://example.com/jdoe.jpg",
        headers=headers,
        content=cover_art,
        status_code=200,
    )
    await person.afetch_avatar()
    assert person.avatar
    assert person.avatar.url


@pytest.mark.parametrize(
    "person_url,img_url",
    [
        ("https://example.com/people/mrbig", None),
        ("https://example.com/people/mrbig", "https://example.com/people/mrbig/me.jpg"),
        (None, "https://example.com/people/mrbig/me.jpg"),
        (None, None),
    ],
)
def test_merge_person(
    mute_signals,
    podcast_with_parsed_episodes,
    podcast_with_two_seasons,
    person_url,
    img_url,
):
    source_person = (
        podcast_with_parsed_episodes.episodes.first().hosts_detected_from_feed.first()
    )
    dest_person = Person.objects.create(
        name="Primary Record", url=person_url, img_url=img_url
    )
    guest_ep = podcast_with_two_seasons.episodes.latest("release_datetime")
    guest_ep.guests_detected_from_feed.add(source_person)
    current_episode_count = source_person.get_total_episodes()
    updated_records = Person.merge_person(source_person, dest_person, dry_run=True)
    assert updated_records == current_episode_count
    assert source_person.get_total_episodes() == current_episode_count
    assert dest_person.get_total_episodes() == 0
    updated_records = Person.merge_person(source_person, dest_person)
    assert updated_records == current_episode_count
    assert source_person.get_total_episodes() == 0
    assert dest_person.get_total_episodes() == current_episode_count
    assert dest_person.guest_appearances.count() == 1
    assert source_person.guest_appearances.count() == 0
    if not img_url:
        assert dest_person.img_url == source_person.img_url
    else:
        assert dest_person.img_url != source_person.img_url
    if not person_url:
        assert dest_person.url == source_person.url
    else:
        assert dest_person.url != source_person.url


def test_merge_person_with_conflicts(podcast_with_parsed_episodes):
    source_person = (
        podcast_with_parsed_episodes.episodes.first().guests_detected_from_feed.first()
    )
    dest_person = Person.objects.create(name="Johnny Appleseed")
    podcast_with_parsed_episodes.episodes.latest(
        "release_datetime"
    ).hosts_detected_from_feed.add(dest_person)
    orig_source_episodes = source_person.get_total_episodes()
    conflict_data = source_person.get_potential_merge_conflicts(dest_person)
    assert conflict_data._common_ids is None
    conflict_data.common_id_list()
    assert conflict_data._common_ids is not None
    updated_records = Person.merge_person(source_person, dest_person)
    assert updated_records == orig_source_episodes
    assert source_person.get_total_episodes() == 0
    assert dest_person.get_total_episodes() == orig_source_episodes


def test_merge_detection_in_episode_parse(podcast_with_parsed_metadata, parsed_rss):
    source_person = Person.objects.create(
        name="John Doe", url="https://somepersonalwebsite.com"
    )
    dest_person = Person.objects.create(
        name="John Doe", url="https://example.com/people/jdoe/"
    )
    Person.merge_person(source_person, dest_person)
    assert podcast_with_parsed_metadata.episodes.count() == 0
    podcast_with_parsed_metadata.update_episodes_from_feed_data(parsed_rss["episodes"])
    for episode in podcast_with_parsed_metadata.episodes.all():
        hosts = episode.hosts_detected_from_feed.all()
        assert dest_person in hosts
        assert source_person not in hosts


def test_season_detection(podcast_with_parsed_metadata, parsed_rss):
    """
    Checks that seasons don't exist for the podcast in initial state
    and that a season is created after episodes are incorporated.
    """
    assert podcast_with_parsed_metadata.seasons.count() == 0
    podcast_with_parsed_metadata.update_episodes_from_feed_data(parsed_rss["episodes"])
    assert podcast_with_parsed_metadata.seasons.count() == 1


def test_no_season_detection(active_podcast):
    assert active_podcast.seasons.count() == 0


def test_transcript_detection(podcast_with_parsed_episodes):
    episode = podcast_with_parsed_episodes.episodes.latest("release_datetime")
    assert episode.transcript_detected
    episode = podcast_with_parsed_episodes.episodes.get(ep_num=1)
    assert episode.transcript_detected


def test_no_transcript(active_podcast):
    episode = active_podcast.episodes.latest("release_datetime")
    assert not episode.transcript_detected


def test_cw_detection(podcast_with_parsed_episodes):
    cw_episodes = podcast_with_parsed_episodes.episodes.filter(cw_present=True)
    assert cw_episodes.count() == 2


def test_skip_items_without_enclosures(podcast_with_parsed_metadata, parsed_rss):
    """
    Under normal circumstances, an episode record should result in an insert, but
    if there are no enclosures it should be skipped. This tests both the atomic
    classmethod and the overall feed generation.
    """
    parsed_rss["episodes"][0]["enclosures"] = []
    assert not Episode.create_or_update_episode_from_feed(
        podcast_with_parsed_metadata, parsed_rss["episodes"][0]
    )
    podcast_with_parsed_metadata.update_episodes_from_feed_data(parsed_rss["episodes"])
    assert podcast_with_parsed_metadata.episodes.count() == 4


def test_episode_duration_property(podcast_with_parsed_episodes):
    expected_duration = timedelta(seconds=147)
    episode = podcast_with_parsed_episodes.episodes.latest("release_datetime")
    assert episode.duration == expected_duration


def test_no_duration_known(active_podcast):
    episode = active_podcast.episodes.latest("release_datetime")
    assert episode.duration is None


@pytest.mark.parametrize(
    "response_code,use_valid_rss,expected_count",
    [(401, True, 0), (404, True, 0), (500, True, 0), (200, False, 0), (200, True, 5)],
)
def test_full_feed_refresh(
    httpx_mock,
    empty_podcast,
    rss_feed_datastream,
    response_code,
    use_valid_rss,
    expected_count,
):
    """
    Tests the `refresh_feed` method.
    """
    if use_valid_rss:
        datastream = rss_feed_datastream
    else:
        with open(
            "tests/data/malformed_podcast_rss_feed.xml",
            "rb",
        ) as f:
            datastream = BytesIO(f.read())
    if response_code != 200:
        httpx_mock.add_response(url=empty_podcast.rss_feed, status_code=response_code)
    else:
        httpx_mock.add_response(
            url=empty_podcast.rss_feed, status_code=200, content=datastream
        )
    assert empty_podcast.refresh_feed() == expected_count


@pytest.mark.parametrize(
    "response_code,url_change_expected", [(301, True), (302, False)]
)
def test_redirected_feed(
    httpx_mock,
    empty_podcast,
    rss_feed_datastream,
    response_code,
    url_change_expected,
):
    """Test that feed data updates to new url when given a permanent redirect."""
    new_url_header = {"Location": "https://example.com/feed.xml"}
    original_url = empty_podcast.rss_feed
    httpx_mock.add_response(
        url=empty_podcast.rss_feed, status_code=response_code, headers=new_url_header
    )
    httpx_mock.add_response(
        url="https://example.com/feed.xml", status_code=200, content=rss_feed_datastream
    )
    empty_podcast.refresh_feed()
    empty_podcast.refresh_from_db()
    if url_change_expected:
        assert empty_podcast.rss_feed == "https://example.com/feed.xml"
    else:
        assert empty_podcast.rss_feed == original_url


def test_analysis_group_feed_count(
    analysis_group,
    podcast_with_parsed_episodes,
    active_podcast,
    podcast_with_two_seasons,
):
    podcast_with_parsed_episodes.analysis_group.add(analysis_group)
    for ep in active_podcast.episodes.all()[:3]:
        ep.analysis_group.add(analysis_group)
    season = podcast_with_two_seasons.seasons.get(season_number=1)
    season.analysis_group.add(analysis_group)
    analysis_group.refresh_from_db()
    assert analysis_group.num_podcasts() == 3
    assert analysis_group.num_seasons() == 2
    assert analysis_group.num_episodes() == 18
    assert analysis_group.num_people() == 2
    season.analysis_group.clear()
    analysis_group.refresh_from_db()
    assert analysis_group.num_podcasts() == 2
    assert analysis_group.num_seasons() == 1
    assert analysis_group.num_episodes() == 8
    assert analysis_group.num_people() == 2
    podcast_with_parsed_episodes.analysis_group.clear()
    analysis_group.refresh_from_db()
    assert analysis_group.num_podcasts() == 1
    assert analysis_group.num_episodes() == 3
    assert analysis_group.num_seasons() == 0
    assert analysis_group.num_people() == 0


def test_analysis_group_categories(
    mute_signals, podcast_with_parsed_episodes, analysis_group
):
    new_podcast = PodcastFactory()
    new_podcast.analysis_group.add(analysis_group)
    podcast_with_parsed_episodes.analysis_group.add(analysis_group)
    categories = analysis_group.get_itunes_categories_with_count()
    assert categories.exists()
    assert categories.count() == 2
    for category in categories:
        assert category.ag_pods == 1
    new_podcast.itunes_categories.set(
        podcast_with_parsed_episodes.itunes_categories.all()
    )
    analysis_group.refresh_from_db()
    categories = analysis_group.get_itunes_categories_with_count()
    assert categories.count() == 2
    for category in categories:
        assert category.ag_pods == 2


def test_analysis_group_calculate_duration_episodes(
    podcast_with_parsed_episodes, analysis_group
):
    podcast_with_parsed_episodes.analysis_group.add(analysis_group)
    assert (
        podcast_with_parsed_episodes.median_episode_duration
        == analysis_group.median_episode_duration
    )
    assert (
        podcast_with_parsed_episodes.total_duration_seconds
        == analysis_group.get_total_duration_seconds()
    )
    assert (
        analysis_group.get_median_duration_timedelta()
        == podcast_with_parsed_episodes.median_episode_duration_timedelta
    )
    assert (
        podcast_with_parsed_episodes.total_duration_timedelta
        == analysis_group.get_total_duration_timedelta()
    )


def test_calculate_median_duration_as_list(podcast_with_parsed_episodes):
    episode_list = list(podcast_with_parsed_episodes.episodes.all())
    assert (
        calculate_median_episode_duration(episode_list)
        == podcast_with_parsed_episodes.median_episode_duration
    )
    assert calculate_median_episode_duration([]) == 0
